"""
Pydantic models for YAKET Trainer.

This package requires the following modules:
    - `pydantic`: https://pydantic-docs.helpmanual.io/
    - `pyyaml`: https://pyyaml.org/wiki/PyYAMLDocumentation
"""
